############ R function: logScrD ##########

# For execution of adaptive trapezoidal approximation
# of the logarithm of the D(p,q,r,s) integrals
# as appearing in McLean and Wand (2019).
# To avoid overflow/underflow, the logarithm is
# returned.

# Last changed: 18 MAY 2020

logScrD <- function(pp,qq,rr,ss,graphChk=FALSE,numBisect=100,
                     suppTolerance=1e-16,convTolerance=1e-5,
                     maxIntgGridSize = NULL)

{
   # Define a function for trapezoidal integration:

   trapezInt <- function(xgrid,fgrid)
   {
      ng <- length(xgrid)
      xvec <- xgrid[2:ng] - xgrid[1:(ng-1)]
      fvec <- fgrid[1:(ng-1)] + fgrid[2:ng]
      return(sum(xvec*fvec)/2)
   }
  
   # Check legality of input parameters:

   if (!any(pp==0:2)) stop("pp must be 0, 1 or 2")
   if (!any(qq==0:2)) stop("qq must be 0, 1 or 2")
   if (rr<0) stop("rr must be non-negative")
   if (ss>=(-rr)) stop("ss must be less than -rr")
     
   # Determine the number of trapezoidal integrations to be done:

   if (pp==0) numTpz <- 1
   if (pp>0) numTpz <- 2

   # Set the exponent functions and required derivatives:

   T1 <- function(x)
   {
      nzi <- (1:length(x))[x!=0]
      ans <- rep(-Inf,length(x))
      ans[nzi] <- x[nzi]*log(x[nzi]) - lgamma(x[nzi])
      return(ans)
   } 

   T1d <- function(x)
   {
      nzi <- (1:length(x))[x!=0]
      ans <- rep(Inf,length(x))
      ans[nzi] <- 1 + logmdigamma(x[nzi])
      return(ans)
   }

   h0 <- function(x,qq,rr,ss)
   {
      ex <- exp(x)
      return((qq+1)*x + rr*T1(ex) + ss*ex)
   }   

   hd <- function(x,pp,qq,rr,ss)
   {
      ex <- exp(x)
      ans0 <- qq + 1 + rr*ex*T1d(ex) + ss*ex
      if (pp==0) return(ans0)
      if (pp>0) return(ans0 + T1d(ex)/(x-(lgamma(ex)/ex)))
   }
     
   hdd <- function(x,pp,qq,rr,ss)
   {
      ex <- exp(x)
      rTerm <- ex*(2 + logmdigamma(ex) - ex*trigamma(ex))

      ans0 <- rr*rTerm + ss*ex

      if (pp==0) return(ans0)
      if (pp>0) 
      {
         pTermNum <- x - 2*(lgamma(ex)/ex) - trigamma(ex)*(x*ex - lgamma(ex)) + digamma(ex) - 1
         pTermDen <- (x - (lgamma(ex)/ex))^2
         pTerm <- sign(pTermNum)*exp(log(abs(pTermNum)) - log(pTermDen))
         return(ans0 + pTerm)
      }
   }   
      
   # Perform bisection search for the roots of hd.
   # For p=0 there is only one root. For p=1 there
   # are two roots.

   mu0 <- rep(NA,numTpz) ; sig0 <- rep(NA,numTpz)

   for (rootSearch in 1:numTpz)
   {
      if (numTpz==2)
      {
         if (rootSearch==1) posFlag <- TRUE
         if (rootSearch==2) posFlag <- FALSE
      }
         
      if (pp==0) {mu0low <- -1 ; mu0upp <- 1}
      if (pp>0)
      {
         if (posFlag) {mu0low <- 0.1 ; mu0upp <- 1}    # Positive axis.
         if (!posFlag) {mu0low <- -1 ; mu0upp <- -0.1} # Negative axis. 
      }

      needBisecSearch <- TRUE
      rootCaptured <- FALSE
      while(!rootCaptured)
      {
         hdLow <- hd(mu0low,pp,qq,rr,ss)
         hdUpp <- hd(mu0upp,pp,qq,rr,ss)
         
         # Take care of cases where root corresponds
         # to hdLow or hdUpp:
         
         if (hdLow==0)
         {
            rootCaptured <- TRUE
            needBisecSearch <- FALSE
            mu0mid <- mu0low
         }

         if (hdUpp==0)
         {  
            rootCaptured <- TRUE
            needBisecSearch <- FALSE
            mu0mid <- mu0upp
         }

         # Now deal with the more usual case
         # where bisection searching is needed:
         
         if (hdLow<0)
         {
            if (pp==0) mu0low <- 2*mu0low
            if (pp>0)
            {
               if (posFlag) mu0low <- mu0low/2
               if (!posFlag) mu0low <- 2*mu0low
            }   
         }
         if (hdUpp>0)
         {
            if (pp==0) mu0upp <- 2*mu0upp
            if (pp>0)
            {
               if (posFlag) mu0upp <- 2*mu0upp
               if (!posFlag) mu0upp <- mu0upp/2
            }   
         }
         if ((hdLow>0)&(hdUpp<0)) rootCaptured <- TRUE
      }

      if (needBisecSearch)
      {
         for (ib in 1:numBisect)
         {
            mu0mid <- (mu0low + mu0upp)/2
            hdLow <- hd(mu0low,pp,qq,rr,ss)
            hdUpp <- hd(mu0upp,pp,qq,rr,ss)
            hdMid <- hd(mu0mid,pp,qq,rr,ss)
            if (hdMid>0) mu0low <- mu0mid
            if (hdMid<0) mu0upp <- mu0mid
         }
      }
      
      # Obtain `best' mean and standard deviation:

      mu0[rootSearch] <- mu0mid
      sig0[rootSearch] <- 1/sqrt(-hdd(mu0[rootSearch],pp,qq,rr,ss))
   }

   # Determine normalising constant to be the one that
   # makes the maximum of all integrals equal to one:

   if (pp==0) mu0max <- mu0[1]

   if (pp>0)
   {
      if (h0(mu0[1],qq,rr,ss) > h0(mu0[2],qq,rr,ss)) mu0max <- mu0[1]
      if (h0(mu0[2],qq,rr,ss) > h0(mu0[1],qq,rr,ss)) mu0max <- mu0[2]
   }
      
   tpzintTotal <- 0
   for (i in 1:numTpz)
   {
      if (numTpz==2)
      {
         if (i==1) posFlag <- TRUE
         if (i==2) posFlag <- FALSE
      }
   
      # Do trapezoidal approximation of suitably
      # translated/dilated integral; starting with
      # search of a good range:

      suppVal <- 4
      suppChosen <- FALSE
   
      while(!suppChosen)
      {
         tailRatioPos <- exp(h0(mu0[i]+suppVal*sig0[i]*sqrt(2),qq,rr,ss)
                            -h0(mu0[i],qq,rr,ss))
         tailRatioNeg <- exp(h0(mu0[i]-suppVal*sig0[i]*sqrt(2),qq,rr,ss)
                             -h0(mu0[i],qq,rr,ss))
         if (pp==0)
            tailRatio <- max(tailRatioPos,tailRatioNeg)

         if (pp>0)
         {
            if (posFlag) tailRatio <- tailRatioPos
            if (!posFlag) tailRatio <- tailRatioNeg
         }
         if (tailRatio<suppTolerance) suppChosen <- TRUE
         if (tailRatio>=suppTolerance) suppVal <- suppVal + 1
      }

      # Obtain initial answer via a relatively coarse trapezoidal
      # evaluation. Note that, to avoid overflow, we work with
      # h0(x) - h0(mu0max).

      ng <- 1000
      if (pp==0) {gridLow <- -suppVal ; gridUpp <- suppVal}
      if (pp>0)
      {
         if (posFlag) {gridLow <- -mu0[i]/(sig0[i]*sqrt(2)) ; gridUpp <- suppVal}
         if (!posFlag) {gridLow <- -suppVal ; gridUpp <- -mu0[i]/(sig0[i]*sqrt(2))}
      }  
      xg <- seq(gridLow,gridUpp,length=(ng+1))
      xArgg <- xg*sig0[i]*sqrt(2) + mu0[i]
      yg <- exp(h0(xArgg,qq,rr,ss) - h0(mu0max,qq,rr,ss))
      if (pp>0) yg <- (T1(exp(xArgg))^pp)*yg
      tpzintOld <- trapezInt(xg,yg)

      # Re-calculate the answer with successive doubling of
      # the grid-size until relative improvement is negligible:
 
      converged <- FALSE
      while (!converged)
      {
         ng <- 2*ng
         if (ng>50000) cat("WARNING: high ng.\n")
         xg <- seq(gridLow,gridUpp,length=(ng+1))
         xArgg <- xg*sig0[i]*sqrt(2) + mu0[i]
         yg <- exp(h0(xArgg,qq,rr,ss) - h0(mu0max,qq,rr,ss))
         if (pp>0) yg <- (T1(exp(xArgg))^pp)*yg
         tpzint <- trapezInt(xg,yg)
         if (tpzintOld!=0) relErr <- abs((tpzint-tpzintOld)/tpzintOld)
         if (tpzintOld==0) relErr <- 0
         if (relErr<convTolerance) converged <- TRUE
         if (relErr>=convTolerance) tpzintOld <- tpzint     
      }   

      if (graphChk)
      {
         cat("Final grid-size is ",ng,"\n")
         plot(xg,yg,type="n",xlab="x",ylab="integrand",bty="l")
         lines(xg,yg,col="blue")
         points(xg,yg,col="red")
         abline(h=0)
         lines(rep(0,2),c(0,max(yg)),col="forestgreen")
         cat("Hit Enter to continue.\n")
         dummyAns <- readline()
      }  
      tpzintTotal <- tpzintTotal + sqrt(2)*sig0[i]*tpzint
   }

   modulusVal <- h0(mu0max,qq,rr,ss) + log(abs(tpzintTotal))
   signVal <- sign(tpzintTotal)

   return(list(modulus=modulusVal,sign=signVal))
}   

############ End of logScrD ############




