########## R script: t2Lev ###########

# For fitting a 2-level model with t responses
# via variational message passing (VMP). 
# The VMP results can be compared with those of
# mean field variational Bayes (MFVB)
# or Markov chain Monte Carlo (MCMC).
# VMP and MFVB produce the same results.

# Last changed: 08 OCT 2021


# Set tasks:

doVMP <- TRUE 
doMFVB <- FALSE 
doMCMC <- TRUE           ;   compileCode <- TRUE  
checkMFVBvsVMP <- FALSE
plotMCMCvsVMP <- TRUE    ;   showAccuracy <- FALSE   ;   doPDF <- FALSE
plotVMPfits <- FALSE

# Set seed:

set.seed(123)

# Set dataset size:

m <- 20
nVec <- rep(15,m)

if (plotVMPfits) layoutVec <- c(5,4)

# Set convergence tolerance of MFVB and VMP:

MFVBTolerance <- 1e-10   ;   VMPTolerance <- 1e-10

# Set maximum number of iterations:

maxIter <- 500

# Load required libraries and functions:

library(MASS)
library(statmod)
library(matrixcalc)

source("GVMP.r")
source("logmdigammaInv.r")
source("logScrD.r")
source("updateGaussPenFrag.r")
source("updateInvGWishPriorFrag.r")
source("updateIterInvGWishFrag.r") 
source("updatetLikFrag.r")
source("vecInverse.r")

# Set true parameter values:

p <- 2
q <- 2

beta0True <- -0.58
beta1True <- 1.89
SigmaTrue <- matrix(c(2.58,0.22,0.22,1.73),2,2)
sigmaEpsTrue <- 0.2
nuTrue <- 1.5

# Set hyperparameters:

mu.beta <- rep(0,p)
sigsq.beta <- 1e10
Sigma.beta <- sigsq.beta*diag(q)
s.sigsq <- 1e5
s.Sigma <- 1e5  
Lambda.Sigma <- 0.5*diag(rep(1/s.Sigma^2,q))
lambda.nu <- 0.01

# Generate data:

n <- sum(nVec)
x <- NULL; y <- NULL
idnum <- NULL; currStt <- 1
reBlockInds <- vector("list",length=m)
ZList <- vector("list",length=m)

u <- mvrnorm(m,rep(0,2),SigmaTrue)

for (i in 1:m) 
{
  idnum <- c(idnum,rep(i,nVec[i]))
  
  xCurr <- runif(nVec[i])
  # muCurr <- beta0True + u[i,1] + (beta1True + u[i,2])*xCurr
  # yCurr <- muCurr + sigmaEpsTrue*rt(nVec[i],nuTrue)
  muCurr <- beta0True + u[i,1] + (beta1True + u[i,2])*xCurr
  bVec <- 1/rgamma(nVec[i],shape=nuTrue/2,rate=nuTrue/2)
  sdVec <- sigmaEpsTrue*sqrt(bVec)
  yCurr <- rnorm(nVec[i],muCurr,sdVec)
  
  x <- c(x,xCurr);  y <- c(y,yCurr)
  
  currEnd <- currStt + length(xCurr) - 1
  reBlockInds[i] <- list(currStt:currEnd)
  currStt <- currEnd + 1
}

# Obtain design matrix C:

X <- cbind(rep(1,n),x)
C <- matrix(0,n,(p+q*m))
C[(1:n),(1:p)] <- X
nSum <- 0
for (i in 1:m)
{
  ZList[[i]] <- X[reBlockInds[[i]],]
  C[((1+nSum):(nSum+nVec[i])),((p+1+(i-1)*q):(p+i*q))] <- ZList[[i]]
  nSum <- nSum + nVec[i]
}

# Perform MCMC via rstan:

if (doMCMC)
{
  library(rstan)
  
  # Select length of burn-in, size of the kept sample 
  # and thinning factor:
  
  nWarm <- 1000    
  nKept <- 5000 
  nThin <- 1
  
  # Specify model in Stan:
  
  t2LevModel <- 
  'data
  {
    int<lower=1> numObs;            int<lower=1> m;
    real<lower=0> sigmaBeta;        real<lower=0> sSigma;
    real<lower=0> ssigsq;             real<lower=0> nuHYP; 
    real y[numObs];                 matrix[numObs,2] X;
    int idnum[numObs];
  }
  transformed data
  {
    vector[2] zeroVec;              zeroVec = rep_vector(0,2);
  }
  parameters
  {
    vector[2] beta;                 vector[2] u[m];
    cov_matrix[2] Sigma; 
    real<lower=0> sigmaEps;         vector[2] aSigma;
    vector<lower=0>[numObs] bVec;
    real<lower=0> nu;
  }
  transformed parameters
  {
    vector[numObs] meanVec;
    vector<lower=0>[numObs] sigmaAdj;

    meanVec = X*beta + to_vector(u[idnum,1])
                + to_vector(u[idnum,2]).*col(X,2);
    
    sigmaAdj = sigmaEps*sqrt(bVec);
  }
  model
  {
    vector[2] scaleSigma;
    y ~ normal(meanVec,sigmaAdj);
    u ~ multi_normal(zeroVec,Sigma);
    beta ~ normal(0,sigmaBeta);  
    bVec ~ inv_gamma(nu/2,nu/2);
    sigmaEps ~ cauchy(0,ssigsq);
    aSigma ~ inv_gamma(0.5,pow(sSigma,-2));
    for (k in 1:2) scaleSigma[k] = 4/aSigma[k];
    Sigma ~ inv_wishart(3,diag_matrix(scaleSigma));
    nu ~ gamma(1,nuHYP); 
  }'
  
  # Set up data list:
  
  allData <- list(y=y,X=X,idnum=idnum,numObs=n,m=m,sSigma=s.Sigma,
                  ssigsq=s.sigsq,nuHYP=lambda.nu,sigmaBeta=sqrt(Sigma.beta[1,1]))
  
  # Compile code for model if required:
  
  if (compileCode)
    stanCompilObj <- stan(model_code=t2LevModel,data=allData,
                          iter=1,chains=1)
  
  # Perform MCMC:
  
  stanObj <-  stan(model_code=t2LevModel,data=allData,warmup=nWarm,
                   iter=(nWarm + nKept),chains=1,thin=nThin,
                   refresh=200,fit=stanCompilObj)
  
  # Save and extract relevant MCMC samples:
  
  betaMCMC <- NULL
  for (i in 1:p)
  {
    charVar <- paste("beta[",as.character(i),"]",sep="")
    betaMCMC <- rbind(betaMCMC,extract(stanObj,charVar,permuted=FALSE))
  }
  
  sigmaMCMC <- as.vector(extract(stanObj,"sigmaEps",permuted=FALSE))
  
  nuMCMC <- as.vector(extract(stanObj,"nu",permuted=FALSE))
  
  u10MCMC <- as.vector(extract(stanObj,"u[1,1]",permuted=FALSE))
  u11MCMC <- as.vector(extract(stanObj,"u[1,2]",permuted=FALSE))
  u20MCMC <- as.vector(extract(stanObj,"u[2,1]",permuted=FALSE))
  u21MCMC <- as.vector(extract(stanObj,"u[2,2]",permuted=FALSE))
  
  Sigma11MCMC <- as.vector(extract(stanObj,"Sigma[1,1]",permuted=FALSE))
  Sigma12MCMC <- as.vector(extract(stanObj,"Sigma[1,2]",permuted=FALSE))
  Sigma22MCMC <- as.vector(extract(stanObj,"Sigma[2,2]",permuted=FALSE))
  sigma1MCMC <- sqrt(Sigma11MCMC)
  sigma2MCMC <- sqrt(Sigma22MCMC)
  rhoMCMC <- Sigma12MCMC/(sigma1MCMC*sigma2MCMC)
  MCsamplesList <- list(sigma1MCMC,sigma2MCMC,rhoMCMC) 
}

if (doMFVB)
{
  tempMat <- matrix(0,p+m*q,p+m*q)
  diag(tempMat[1:p,1:p]) <- 1/sigsq.beta
  
  mu.q.recip.sigsq.MFVB <- 1
  mu.q.recip.a.MFVB <- 1
  mu.q.recip.b.MFVB <- rep(1,n) 
  M.q.inv.Sigma.MFVB <- diag(q)
  M.q.inv.A.MFVB <- diag(q)
  
  nu.sigsq <- 1   ;   nu.Sigma <- 2
  delta.q.sigsq.MFVB <- nu.sigsq + sum(nVec)
  delta.q.a.MFVB <- nu.sigsq + 1
  xi.q.Sigma.MFVB <- nu.Sigma + 2*q - 2 + m
  xi.q.A.MFVB <- nu.Sigma + q
  
  Lambda.q.A.MFVB <- matrix(0, q, q)
  
  mu.q.nu.MFVB <- 1
  
  # Perform MFVB:
  
  converged <- FALSE  
  itNum <- 0
  while (!converged)
  {
    itNum <- itNum + 1 
    
    # Do updates for q*(beta,u):
    
    Cmuqrecipb <- C*mu.q.recip.b.MFVB
    CTmuqrecipbC <- crossprod(Cmuqrecipb,C)
    tempMat[-(1:p),-(1:p)] <- kronecker(diag(m),M.q.inv.Sigma.MFVB)
    Sigma.q.betau.MFVB <- solve(mu.q.recip.sigsq.MFVB*CTmuqrecipbC + tempMat)
    mu.q.betau.MFVB <-  mu.q.recip.sigsq.MFVB*crossprod(Sigma.q.betau.MFVB,crossprod(Cmuqrecipb,y))
    
    resVec <- as.vector(y - C%*%mu.q.betau.MFVB)

    # Do updates for q*(sigma^2):
    
    lambda.q.sigsq.MFVB <- (mu.q.recip.a.MFVB + sum(mu.q.recip.b.MFVB*(resVec*resVec
                            +  diag(tcrossprod(tcrossprod(C,Sigma.q.betau.MFVB),C)))))
    mu.q.recip.sigsq.MFVB <- delta.q.sigsq.MFVB/lambda.q.sigsq.MFVB
    
    # Do updates for q*(Sigma):
    
    Lambda.q.Sigma.MFVB <- M.q.inv.A.MFVB
    for (i in 1:m)
    {
      sttVal <- p+q*(i-1) + 1  ;  endVal <- p + q*i
      Lambda.q.Sigma.MFVB <- Lambda.q.Sigma.MFVB + (tcrossprod(mu.q.betau.MFVB[sttVal:endVal]) 
                              + Sigma.q.betau.MFVB[sttVal:endVal,sttVal:endVal])
    }

    M.q.inv.Sigma.MFVB <- (xi.q.Sigma.MFVB - q + 1)*solve(Lambda.q.Sigma.MFVB)

    # Do updates for q*(a):
    
    lambda.q.a.MFVB <- mu.q.recip.sigsq.MFVB + (1/(nu.sigsq*s.sigsq^2))
    mu.q.recip.a.MFVB <- delta.q.a.MFVB/lambda.q.a.MFVB
    
    # Do updates for q*(A):
    
    diag(Lambda.q.A.MFVB) <- diag(M.q.inv.Sigma.MFVB) +  (1/(nu.Sigma*s.Sigma^2))
    M.q.inv.A.MFVB <- xi.q.A.MFVB*solve(Lambda.q.A.MFVB)
    
    # Do updates for q*(b):
    
    delta.q.b.MFVB <- mu.q.nu.MFVB + 1
    lambda.q.b.MFVB <- mu.q.nu.MFVB + mu.q.recip.sigsq.MFVB*(resVec*resVec
                        + diag(tcrossprod(tcrossprod(C,Sigma.q.betau.MFVB),C)))
    mu.q.recip.b.MFVB <- as.vector(delta.q.b.MFVB/lambda.q.b.MFVB)
    mu.q.log.b.MFVB <- log(0.5*lambda.q.b.MFVB) - digamma(0.5*delta.q.b.MFVB)
    
    # Do updates for q*(nu):
    
    beta.q.nu.MFVB <- sum(mu.q.log.b.MFVB + mu.q.recip.b.MFVB) + lambda.nu
    mu.q.nu.MFVB <- 2*exp(logScrD(0,1,n,-beta.q.nu.MFVB)$modulus
                          - logScrD(0,0,n,-beta.q.nu.MFVB)$modulus)
    
    # Check for convergence:
    
    parVecHat <- c(mu.q.betau.MFVB,vech(Sigma.q.betau.MFVB),
                   mu.q.recip.sigsq.MFVB,mu.q.recip.a.MFVB,
                   vech(M.q.inv.Sigma.MFVB),vech(M.q.inv.A.MFVB),
                   mu.q.recip.b.MFVB,mu.q.log.b.MFVB,mu.q.nu.MFVB)
    
    if (itNum>1)
    {
      relErrTemp <- abs((parVecHat - parVecHatPrev)/parVecHatPrev)
      relErrTemp[is.nan(relErrTemp)] <- 0
      relErr <- max(relErrTemp)
      cat("Iteration number:",itNum,";","MFVB relative error:",signif(relErr,5),"\n")
      if (relErr<MFVBTolerance && relErr!=0)
        converged <- TRUE
    }
    if (itNum>=maxIter)
    {
      converged <- TRUE
      warning("Maximum number of MFVB iterations exceeded.\n")
    }
    
    if (itNum>=maxIter) converged <- TRUE
    
    parVecHatPrev <- parVecHat
  }
}

if (doVMP)
{
  # Obtain dusplication matrix and its Moore-Penrose inverse:

  Dq <- duplication.matrix(q)  ;   DqPlus <- solve(crossprod(Dq),t(Dq))
  
  # Set-up lists named "eta" and "G" with component names 
  # corresponding to the messages:
  
  eta <- vector("list",28)
  G <- vector("list",8)
  
  # Initialise a and A graphs:
  
  Ga <- GA <- "Gdiag"
  
  # Initialise factor to stochastic node natural parameters and graphs:
  
  G$"p(sigsq|a)->a" <- G$"p(Sigma|A)->A" <- "Gdiag"
  G$"p(sigsq|a)->sigsq" <- G$"p(Sigma|A)->Sigma" <- "Gfull"
  aOut <- updateInvGWishPriorFrag(Ga,1,1/s.sigsq^2)
  eta$"p(a)->a" <- aOut$eta   ;   G$"p(a)->a" <- aOut$G
  Aout <- updateInvGWishPriorFrag(GA,1,Lambda.Sigma)
  eta$"p(A)->A" <- Aout$eta   ;   G$"p(A)->A" <- Aout$G
  eta$"p(sigsq|a)->a" <-  c(-2,-1)   
  eta$"p(Sigma|A)->A" <- c(-0.5,-0.5*crossprod(Dq,vec(diag(q))))
  eta$"p(ups)->ups" <- c(0,(-lambda.nu))
  eta$"p(b|ups)->ups" <- c(1,-1.1)
  eta$"p(beta,u|Sigma)->Sigma" <- c(-0.5,-0.5*vec(diag(q)))
  eta$"p(Sigma|A)->Sigma" <- eta$"p(Sigma|A)->A"
  eta$"p(y|beta,u,sigsq,b)->sigsq" <- c(-2,-1) 
  eta$"p(sigsq|a)->sigsq" <- c(-2,-1)
  eta$"p(y|beta,u,sigsq,b)->(beta,u)" <- c(rep(0,p+m*q),-0.5*vec(diag(p+m*q)))
  eta$"p(beta,u|Sigma)->(beta,u)" <- eta$"p(y|beta,u,sigsq,b)->(beta,u)"

  # Initialise stochastic node to factor message natural parameters and graphs:

  Ga <- GA <- "Gfull"
  G$"a->p(sigsq|a)" <- G$"p(a)->a"   ;   G$"A->p(Sigma|A)" <- G$"p(A)->A"
  eta$"a->p(sigsq|a)" <- eta$"p(a)->a" 
  eta$"A->p(Sigma|A)" <- eta$"p(A)->A"
  eta$"ups->p(b|ups)" <- eta$"p(ups)->ups"
  
  # Perform VMP:

  converged <- FALSE  
  itNum <- 0
  while (!converged)
  {
    itNum <- itNum + 1   
    
    # Update message natural parameters for the iterated Inverse G-Wishart
    # fragment dependent on sigma^2:
    
    eta$"sigsq->p(sigsq|a)" <- eta$"p(y|beta,u,sigsq,b)->sigsq" 
    
    GetaIN <- list(G=G$"a->p(sigsq|a)",
                   eta=list(eta$"sigsq->p(sigsq|a)",eta$"p(sigsq|a)->sigsq",
                            eta$"a->p(sigsq|a)",eta$"p(sigsq|a)->a"))
      
    GetaOUT <- updateIterInvGWishFrag(1,Ga,GetaIN)
    
    G$"p(sigsq|a)->sigsq" <- GetaOUT$G[[1]]
    G$"p(sigsq|a)->a" <- GetaOUT$G[[2]]
    eta$"p(sigsq|a)->sigsq" <- GetaOUT$eta[[1]]
    eta$"p(sigsq|a)->a" <- GetaOUT$eta[[2]]
    
    # Update message natural parameters for the iterated Inverse G-Wishart
    # fragment dependent on Sigma:
    
    eta$"Sigma->p(Sigma|A)" <- c(eta$"p(beta,u|Sigma)->Sigma"[1],
                                 crossprod(Dq,eta$"p(beta,u|Sigma)->Sigma"[-1]))
    
    GetaIN <- list(G=G$"A->p(Sigma|A)",
                   eta=list(eta$"Sigma->p(Sigma|A)",eta$"p(Sigma|A)->Sigma",
                            eta$"A->p(Sigma|A)",eta$"p(Sigma|A)->A"))
    
    GetaOUT <- updateIterInvGWishFrag(2*q,GA,GetaIN)
    
    G$"p(Sigma|A)->Sigma" <- GetaOUT$G[[1]]
    G$"p(Sigma|A)->A" <- GetaOUT$G[[2]]
    eta$"p(Sigma|A)->Sigma" <- GetaOUT$eta[[1]]
    eta$"p(Sigma|A)->A" <- GetaOUT$eta[[2]]
    
    # Update message natural parameters for Gaussian penalisation fragment:
    
    eta$"(beta,u)->p(beta,u|Sigma)" <- eta$"p(y|beta,u,sigsq,b)->(beta,u)"
    eta$"Sigma->p(beta,u|Sigma)" <- c(eta$"p(Sigma|A)->Sigma"[1],
                                      crossprod(DqPlus,eta$"p(Sigma|A)->Sigma"[-1]))
    
    etaIN <- list(eta$"(beta,u)->p(beta,u|Sigma)",eta$"p(beta,u|Sigma)->(beta,u)",
                  eta$"Sigma->p(beta,u|Sigma)",eta$"p(beta,u|Sigma)->Sigma")
    
    etaOUT  <- updateGaussPenFrag(mu.beta,Sigma.beta,m*q,etaIN)
    
    eta$"p(beta,u|Sigma)->(beta,u)" <- etaOUT[[1]]
    eta$"p(beta,u|Sigma)->Sigma" <- etaOUT[[2]]

    # Update message natural parameters for t likelihood fragment:
    
    eta$"(beta,u)->p(y|beta,u,sigsq,b)" <- eta$"p(beta,u|Sigma)->(beta,u)"
    eta$"sigsq->p(y|beta,u,sigsq,b)" <- eta$"p(sigsq|a)->sigsq"
    
    etaIN <- list(eta$"(beta,u)->p(y|beta,u,sigsq,b)",eta$"p(y|beta,u,sigsq,b)->(beta,u)",
                  eta$"sigsq->p(y|beta,u,sigsq,b)",eta$"p(y|beta,u,sigsq,b)->sigsq",
                  eta$"ups->p(b|ups)",eta$"p(b|ups)->ups")
    
    etaOUT  <- updatetLikFrag(y,C,etaIN)
    
    eta$"p(y|beta,u,sigsq,b)->(beta,u)" <-  etaOUT[[1]]
    eta$"p(y|beta,u,sigsq,b)->sigsq" <- etaOUT[[2]]
    eta$"p(b|ups)->ups" <- etaOUT[[3]]
    
    # Check for convergence:
    
    parVecHat <- c(eta$"p(sigsq|a)->sigsq",eta$"p(sigsq|a)->a",
                   eta$"p(Sigma|A)->Sigma",eta$"p(Sigma|A)->A",
                   eta$"p(beta,u|Sigma)->(beta,u)",eta$"p(beta,u|Sigma)->Sigma",
                   eta$"p(y|beta,u,sigsq,b)->(beta,u)",eta$"p(y|beta,u,sigsq,b)->sigsq",
                   eta$"p(b|ups)->ups")
    
    if (itNum>1)
    {
      relErrTemp <- abs((parVecHat - parVecHatPrev)/parVecHatPrev)
      relErrTemp[is.nan(relErrTemp)] <- 0
      relErr <- max(relErrTemp)
      cat("Iteration number:",itNum,";","VMP relative error:",signif(relErr,5),"\n")
      if (relErr<VMPTolerance && relErr!=0)
        converged <- TRUE
    }
    if (itNum>=maxIter)
    {
      converged <- TRUE
      warning("Maximum number of VMP iterations exceeded.\n")
    }
    
    if (itNum>=maxIter) converged <- TRUE
    
    parVecHatPrev <- parVecHat
  }
  
  # Obtain q-density natural parameters:
  
  eta$"q(beta,u)" <-  eta$"p(y|beta,u,sigsq,b)->(beta,u)" + eta$"p(beta,u|Sigma)->(beta,u)"
  eta$"q(sigsq)" <- eta$"p(y|beta,u,sigsq,b)->sigsq" + eta$"p(sigsq|a)->sigsq"
  eta$"q(Sigma)" <- (eta$"p(beta,u|Sigma)->Sigma" + 
                     c(eta$"p(Sigma|A)->Sigma"[1],crossprod(DqPlus,eta$"p(Sigma|A)->Sigma"[-1])))
  eta$"q(ups)" <- eta$"p(ups)->ups" + eta$"p(b|ups)->ups"
  
  # Obtain q-density common parameters:
  
  eta1 <- eta$"q(beta,u)"[1:(p+m*q)]  ;  eta2 <- eta$"q(beta,u)"[-(1:(p+m*q))] 
  Sigma.q.betau.VMP <- -0.5*solve(vecInverse(eta2))
  mu.q.betau.VMP <- -0.5*solve(vecInverse(eta2),eta1)
  
  xi.q.Sigma.VMP <- -2 -2*eta$"q(Sigma)"[1]
  Lambda.q.Sigma.VMP <- -2*vecInverse(eta$"q(Sigma)"[-1])
  VMPparList <- list(xi.q.Sigma.VMP,Lambda.q.Sigma.VMP) 
  
  delta.q.sigsq.VMP <- -2 -2*eta$"q(sigsq)"[1]
  lambda.q.sigsq.VMP <- -2*eta$"q(sigsq)"[2]
  
  mu.q.nu.VMP <- 2*exp(logScrD(0,1,eta$"q(ups)"[1],eta$"q(ups)"[2])$modulus
                       -logScrD(0,0,eta$"q(ups)"[1],eta$"q(ups)"[2])$modulus)
  alpha.q.ups.VMP <- eta$"q(ups)"[1]
  beta.q.ups.VMP <- -eta$"q(ups)"[2]
}

if (checkMFVBvsVMP)
{
  M.q.inv.Sigma.VMP <- (xi.q.Sigma.VMP-q+1)*solve(Lambda.q.Sigma.VMP)
  
  outTable <- rbind(cbind(mu.q.betau.MFVB[1:p],mu.q.betau.VMP[1:p]),
                    c(mu.q.recip.sigsq.MFVB,delta.q.sigsq.VMP/lambda.q.sigsq.VMP),
                    c(M.q.inv.Sigma.MFVB[1,1],M.q.inv.Sigma.VMP[1,1]),
                    c(M.q.inv.Sigma.MFVB[2,2],M.q.inv.Sigma.VMP[2,2]),
                    c(M.q.inv.Sigma.MFVB[1,2],M.q.inv.Sigma.VMP[1,2]),
                    c(mu.q.nu.MFVB,mu.q.nu.VMP))
  
  outTable <- as.data.frame(outTable)
  outTable <- cbind(outTable,outTable[,1]/outTable[,2])
  
  dimnames(outTable)[[1]] <- c("mu.q.beta[0]","mu.q.beta[1]","mu.q.recip.sigsq",
                               "M.q.Sigma^{-1}[1,1]","M.q.Sigma^{-1}[2,2]",
                               "M.q.Sigma^{-1}[1,2]","mu.q.nu")
  dimnames(outTable)[[2]] <- c("MFVB","VMP","ratio")
  
  cat("\n\n")
  cat("Summary of selected parameter comparisons:\n\n")
  print(outTable)
  cat("\n\n")
}

if (plotMCMCvsVMP)
{
  # Load required libraries:
  
  library(KernSmooth)
  library(MCMCpack)
  
  if(doPDF) pdf("postDensPlots.pdf",13,9)
  
  # Set framework to draw posterior density plots:
  
  par(mfrow=c(3,4),mai=c(0.5,0.55,0.4,0.1))
  gridSize <- 1001        ;   lwdVal <- 3
  cex.mainVal <- 3        ;   cex.labVal <- 2
  cex.axisVal <- 2
  colMC <- "grey65"         ;   colVA <- "black" 
  ltyMC <- 1              ;   ltyVA <- 2               
  cex.accur <- 2.2        ;   colAcc <- "black"
  colTruth <- "black"
  
  # Define a function for trapezoidal integration:
  
  trapint <- function(xgrid, fgrid) 
  {
    ng <- length(xgrid)
    xvec <- xgrid[2:ng] - xgrid[1:(ng - 1)]
    fvec <- fgrid[1:(ng - 1)] + fgrid[2:ng]
    integ <- sum(xvec * fvec)/2
    return(integ)
  }
  
  ### Plot for beta_0 ###
  
  # Obtain density estimates and accuracy scores:
  
  VAlower <- mu.q.betau.VMP[1] - qnorm(0.9995)*sqrt(Sigma.q.betau.VMP[1,1])
  VAupper <- mu.q.betau.VMP[1] + qnorm(0.9995)*sqrt(Sigma.q.betau.VMP[1,1])
  MClower <- quantile(betaMCMC[1,],0.0005)
  MCupper <- quantile(betaMCMC[1,],0.9995)
  xgLow <- min(MClower,VAlower)   ;   xgUpp <- max(MCupper,VAupper)
  xg <- seq(xgLow,xgUpp,length=gridSize)
  VApostg <- dnorm(xg,mu.q.betau.VMP[1],sqrt(Sigma.q.betau.VMP[1,1]))
  MCpostg <- bkde(betaMCMC[1,],bandwidth=dpik(betaMCMC[1,]),
                  range.x=range(xg),gridsize=gridSize)$y
  accurVal <- round(100-50*trapint(xg,abs(MCpostg-VApostg)),2)
  
  # Obtain approximate posterior plots:
  
  xlimVal <- c(xgLow,xgUpp)
  ylimVal <- range(c(VApostg,MCpostg))
  
  plot(xg,VApostg,type="n",bty="l",main=expression(paste(beta[0])),
       cex.main=cex.mainVal,xlab="",ylab="approx. posterior",
       cex.lab=cex.labVal,cex.axis=cex.axisVal,ylim=ylimVal)
  lines(xg,MCpostg,lwd=lwdVal,col=colMC,lty=ltyMC)
  lines(xg,VApostg,lwd=lwdVal,col=colVA,lty=ltyVA)
  
  if(showAccuracy)
  {
    # Plot accuracy value:
    
    text(0.15*min(xlimVal)+0.85*max(xlimVal),0.1*min(ylimVal)+0.9*max(ylimVal),
         as.character(paste(accurVal,"%",sep="")),cex=cex.accur,col=colAcc)
    text(0.15*min(xlimVal)+0.85*max(xlimVal),0.25*min(ylimVal)+0.75*max(ylimVal),
         "accuracy",cex=0.5*cex.accur,col=colAcc)
  }
  
  # Add true value:
  
  lines(rep(beta0True,2),c(0,ylimVal[2]),col=colTruth)
  abline(h=0)
  
  ### Plot for beta_1 ###
  
  # Obtain density estimates and accuracy scores:
  VAlower <- mu.q.betau.VMP[2] - qnorm(0.9995)*sqrt(Sigma.q.betau.VMP[2,2])
  VAupper <- mu.q.betau.VMP[2] + qnorm(0.9995)*sqrt(Sigma.q.betau.VMP[2,2])
  MClower <- quantile(betaMCMC[2,],0.0005)
  MCupper <- quantile(betaMCMC[2,],0.9995)
  xgLow <- min(MClower,VAlower)   ;   xgUpp <- max(MCupper,VAupper)
  xg <- seq(xgLow,xgUpp,length=gridSize)
  VApostg <- dnorm(xg,mu.q.betau.VMP[2],sqrt(Sigma.q.betau.VMP[2,2]))
  MCpostg <- bkde(betaMCMC[2,],bandwidth=dpik(betaMCMC[2,]),
                  range.x=range(xg),gridsize=gridSize)$y
  accurVal <- round(100-50*trapint(xg,abs(MCpostg-VApostg)),2)
  
  # Obtain approximate posterior plots:
  
  xlimVal <- c(xgLow,xgUpp)
  ylimVal <- range(c(VApostg,MCpostg))
  
  plot(xg,VApostg,type="n",bty="l",main=expression(paste(beta[1])),
       cex.main=cex.mainVal,xlab="",ylab="approx. posterior",
       cex.lab=cex.labVal,cex.axis=cex.axisVal,ylim=ylimVal)
  lines(xg,MCpostg,lwd=lwdVal,col=colMC,lty=ltyMC)
  lines(xg,VApostg,lwd=lwdVal,col=colVA,lty=ltyVA)
  
  if(showAccuracy)
  {
    # Plot accuracy value:
    
    text(0.15*min(xlimVal)+0.85*max(xlimVal),0.1*min(ylimVal)+0.9*max(ylimVal),
         as.character(paste(accurVal,"%",sep="")),cex=cex.accur,col=colAcc)
    text(0.15*min(xlimVal)+0.85*max(xlimVal),0.25*min(ylimVal)+0.75*max(ylimVal),
         "accuracy",cex=0.5*cex.accur,col=colAcc)
  }
  
  # Add true value:
  
  lines(rep(beta1True,2),c(0,ylimVal[2]),col=colTruth)
  abline(h=0)
  
  ### Plot for sigma ###
  
  # Obtain density estimates and accuracy scores:
  
  nMC <- 10000
  sigmaVAsamp <- sqrt(1/rgamma(nMC,0.5*delta.q.sigsq.VMP,rate=0.5*lambda.q.sigsq.VMP))
  xgLow <- quantile(c(sigmaVAsamp,sigmaMCMC),0.0001)
  xgUpp <- quantile(c(sigmaVAsamp,sigmaMCMC),0.9999)
  xg <- seq(xgLow,xgUpp,length=gridSize)
  destVA <- bkde(sigmaVAsamp,bandwidth=dpik(sigmaVAsamp),
                 range.x=range(xg),gridsize=gridSize)
  VApostg <- destVA$y
  MCpostg <- bkde(sigmaMCMC,bandwidth=dpik(sigmaMCMC),
                  range.x=range(xg),gridsize=gridSize)$y
  accurVal <- round(100-50*trapint(xg,abs(MCpostg-VApostg)),2)
  
  # Obtain approximate posterior plots:
  
  xlimVal <- c(xgLow,xgUpp)
  ylimVal <- range(c(VApostg,MCpostg))
  
  plot(xg,VApostg,type="n",bty="l",main=expression(paste(sigma)),
       cex.main=cex.mainVal,xlab="",ylab="approx. posterior",
       cex.lab=cex.labVal,cex.axis=cex.axisVal,ylim=ylimVal)
  lines(xg,MCpostg,lwd=lwdVal,col=colMC,lty=ltyMC)
  lines(xg,VApostg,lwd=lwdVal,col=colVA,lty=ltyVA)
  
  if(showAccuracy)
  {
    # Plot accuracy value:
    
    text(0.15*min(xlimVal)+0.85*max(xlimVal),0.1*min(ylimVal)+0.9*max(ylimVal),
         as.character(paste(accurVal,"%",sep="")),cex=cex.accur,col=colAcc)
    text(0.15*min(xlimVal)+0.85*max(xlimVal),0.25*min(ylimVal)+0.75*max(ylimVal),
         "accuracy",cex=0.5*cex.accur,col=colAcc)
  }
  
  # Add true value:
  
  lines(rep(sigmaEpsTrue,2),c(0,ylimVal[2]),col=colTruth)
  abline(h=0)
  
  ### Plot for nu ###
  
  # Obtain density estimates and accuracy scores:
  
  VAlower <- max(mu.q.nu.VMP-2.5*sd(nuMCMC),0.0001)
  VAupper <- mu.q.nu.VMP+2.5*sd(nuMCMC)
  MClower <- quantile(nuMCMC,0.0005)
  MCupper <- quantile(nuMCMC,0.9995)
  xgLow <- min(MClower,VAlower)   ;   xgUpp <- max(MCupper,VAupper)
  xg <- seq(xgLow,xgUpp,length=gridSize)
  logDenom <- logScrD(0,0,alpha.q.ups.VMP,-beta.q.ups.VMP)$modulus
  logNumer <- (alpha.q.ups.VMP*((xg/2)*log(xg/2) - lgamma(xg/2)) - beta.q.ups.VMP*xg/2)
  VApostg <- exp(logNumer - logDenom)/2
  MCpostg <- bkde(nuMCMC,bandwidth=dpik(nuMCMC),
                  range.x=range(xg),gridsize=gridSize)$y
  accurVal <- round(100-50*trapint(xg,abs(MCpostg-VApostg)),2)
  
  # Obtain approximate posterior plots:
  
  xlimVal <- c(xgLow,xgUpp)
  ylimVal <- range(c(VApostg,MCpostg))
  
  plot(xg,VApostg,type="n",bty="l",main=expression(paste(nu)),
       cex.main=cex.mainVal,xlab="",ylab="approx. posterior",
       cex.lab=cex.labVal,cex.axis=cex.axisVal,ylim=ylimVal)
  lines(xg,MCpostg,lwd=lwdVal,col=colMC,lty=ltyMC)
  lines(xg,VApostg,lwd=lwdVal,col=colVA,lty=ltyVA)
  
  if(showAccuracy)
  {
    # Plot accuracy value:
    
    text(0.15*min(xlimVal)+0.85*max(xlimVal),0.1*min(ylimVal)+0.9*max(ylimVal),
         as.character(paste(accurVal,"%",sep="")),cex=cex.accur,col=colAcc)
    text(0.15*min(xlimVal)+0.85*max(xlimVal),0.25*min(ylimVal)+0.75*max(ylimVal),
         "accuracy",cex=0.5*cex.accur,col=colAcc)
  }

  # Add true value:
  
  lines(rep(nuTrue,2),c(0,ylimVal[2]),col=colTruth)
  abline(h=0)
  
  ### Plot for u_10 ###
  
  # Obtain density estimates and accuracy scores:
  
  VAlower <- mu.q.betau.VMP[p+1] - qnorm(0.9995)*sqrt(Sigma.q.betau.VMP[(p+1),(p+1)])
  VAupper <- mu.q.betau.VMP[p+1] + qnorm(0.9995)*sqrt(Sigma.q.betau.VMP[(p+1),(p+1)])
  MClower <- quantile(u10MCMC,0.0005)
  MCupper <- quantile(u10MCMC,0.9995)
  xgLow <- min(MClower,VAlower)   ;   xgUpp <- max(MCupper,VAupper)
  xg <- seq(xgLow,xgUpp,length=gridSize)
  VApostg <- dnorm(xg,mu.q.betau.VMP[p+1],sqrt(Sigma.q.betau.VMP[(p+1),(p+1)]))
  MCpostg <- bkde(u10MCMC,bandwidth=dpik(u10MCMC),
                  range.x=range(xg),gridsize=gridSize)$y
  accurVal <- round(100-50*trapint(xg,abs(MCpostg-VApostg)),2)
  
  # Obtain approximate posterior plots:
  
  xlimVal <- c(xgLow,xgUpp)
  ylimVal <- range(c(VApostg,MCpostg))
  
  plot(xg,VApostg,type="n",bty="l",main=expression(paste(u[10])),
       cex.main=cex.mainVal,xlab="",ylab="approx. posterior",
       cex.lab=cex.labVal,cex.axis=cex.axisVal,ylim=ylimVal)
  lines(xg,MCpostg,lwd=lwdVal,col=colMC,lty=ltyMC)
  lines(xg,VApostg,lwd=lwdVal,col=colVA,lty=ltyVA)
  
  if(showAccuracy)
  {
    # Plot accuracy value:
    
    text(0.15*min(xlimVal)+0.85*max(xlimVal),0.1*min(ylimVal)+0.9*max(ylimVal),
         as.character(paste(accurVal,"%",sep="")),cex=cex.accur,col=colAcc)
    text(0.15*min(xlimVal)+0.85*max(xlimVal),0.25*min(ylimVal)+0.75*max(ylimVal),
         "accuracy",cex=0.5*cex.accur,col=colAcc)
  }
  
  # Add true value:
  
  lines(rep(u[1,1],2),c(0,ylimVal[2]),col=colTruth)
  abline(h=0)
  
  ### Plot for u_11 ###
  
  # Obtain density estimates and accuracy scores:
  
  VAlower <- mu.q.betau.VMP[p+2] - qnorm(0.9995)*sqrt(Sigma.q.betau.VMP[(p+2),(p+2)])
  VAupper <- mu.q.betau.VMP[p+2] + qnorm(0.9995)*sqrt(Sigma.q.betau.VMP[(p+2),(p+2)])
  MClower <- quantile(u11MCMC,0.0005)
  MCupper <- quantile(u11MCMC,0.9995)
  xgLow <- min(MClower,VAlower)   ;   xgUpp <- max(MCupper,VAupper)
  xg <- seq(xgLow,xgUpp,length=gridSize)
  VApostg <- dnorm(xg,mu.q.betau.VMP[p+2],sqrt(Sigma.q.betau.VMP[(p+2),(p+2)]))
  MCpostg <- bkde(u11MCMC,bandwidth=dpik(u11MCMC),
                  range.x=range(xg),gridsize=gridSize)$y
  accurVal <- round(100-50*trapint(xg,abs(MCpostg-VApostg)),2)
  
  # Obtain approximate posterior plots:
  
  xlimVal <- c(xgLow,xgUpp)
  ylimVal <- range(c(VApostg,MCpostg))
  
  plot(xg,VApostg,type="n",bty="l",main=expression(paste(u[11])),
       cex.main=cex.mainVal,xlab="",ylab="approx. posterior",
       cex.lab=cex.labVal,cex.axis=cex.axisVal,ylim=ylimVal)
  lines(xg,MCpostg,lwd=lwdVal,col=colMC,lty=ltyMC)
  lines(xg,VApostg,lwd=lwdVal,col=colVA,lty=ltyVA)
  
  if(showAccuracy)
  {
    # Plot accuracy value:
    
    text(0.15*min(xlimVal)+0.85*max(xlimVal),0.1*min(ylimVal)+0.9*max(ylimVal),
         as.character(paste(accurVal,"%",sep="")),cex=cex.accur,col=colAcc)
    text(0.15*min(xlimVal)+0.85*max(xlimVal),0.25*min(ylimVal)+0.75*max(ylimVal),
         "accuracy",cex=0.5*cex.accur,col=colAcc) 
  }
  
  # Add true value:
  
  lines(rep(u[1,2],2),c(0,ylimVal[2]),col=colTruth)
  abline(h=0)
  
  ### Plot for u_20 ###
  
  # Obtain density estimates and accuracy scores:
  
  VAlower <- mu.q.betau.VMP[p+3] - qnorm(0.9995)*sqrt(Sigma.q.betau.VMP[(p+3),(p+3)])
  VAupper <- mu.q.betau.VMP[p+3] + qnorm(0.9995)*sqrt(Sigma.q.betau.VMP[(p+3),(p+3)])
  MClower <- quantile(u20MCMC,0.0005)
  MCupper <- quantile(u20MCMC,0.9995)
  xgLow <- min(MClower,VAlower)   ;   xgUpp <- max(MCupper,VAupper)
  xg <- seq(xgLow,xgUpp,length=gridSize)
  VApostg <- dnorm(xg,mu.q.betau.VMP[p+3],sqrt(Sigma.q.betau.VMP[(p+3),(p+3)]))
  MCpostg <- bkde(u20MCMC,bandwidth=dpik(u20MCMC),
                  range.x=range(xg),gridsize=gridSize)$y
  accurVal <- round(100-50*trapint(xg,abs(MCpostg-VApostg)),2)
  
  # Obtain approximate posterior plots:
  
  xlimVal <- c(xgLow,xgUpp)
  ylimVal <- range(c(VApostg,MCpostg))
  
  plot(xg,VApostg,type="n",bty="l",main=expression(paste(u[20])),
       cex.main=cex.mainVal,xlab="",ylab="approx. posterior",
       cex.lab=cex.labVal,cex.axis=cex.axisVal,ylim=ylimVal)
  lines(xg,MCpostg,lwd=lwdVal,col=colMC,lty=ltyMC)
  lines(xg,VApostg,lwd=lwdVal,col=colVA,lty=ltyVA)
  
  if(showAccuracy)
  {
    # Plot accuracy value:
    
    text(0.15*min(xlimVal)+0.85*max(xlimVal),0.1*min(ylimVal)+0.9*max(ylimVal),
         as.character(paste(accurVal,"%",sep="")),cex=cex.accur,col=colAcc)
    text(0.15*min(xlimVal)+0.85*max(xlimVal),0.25*min(ylimVal)+0.75*max(ylimVal),
         "accuracy",cex=0.5*cex.accur,col=colAcc)
  }
  
  # Add true value:
  
  lines(rep(u[2,1],2),c(0,ylimVal[2]),col=colTruth)
  abline(h=0)
  
  ### Plot for u_21 ###
  
  # Obtain density estimates and accuracy scores:
  
  VAlower <- mu.q.betau.VMP[p+4] - qnorm(0.9995)*sqrt(Sigma.q.betau.VMP[(p+4),(p+4)])
  VAupper <- mu.q.betau.VMP[p+4] + qnorm(0.9995)*sqrt(Sigma.q.betau.VMP[(p+4),(p+4)])
  MClower <- quantile(u21MCMC,0.0005)
  MCupper <- quantile(u21MCMC,0.9995)
  xgLow <- min(MClower,VAlower)   ;   xgUpp <- max(MCupper,VAupper)
  xg <- seq(xgLow,xgUpp,length=gridSize)
  VApostg <- dnorm(xg,mu.q.betau.VMP[p+4],sqrt(Sigma.q.betau.VMP[(p+4),(p+4)]))
  MCpostg <- bkde(u21MCMC,bandwidth=dpik(u21MCMC),
                  range.x=range(xg),gridsize=gridSize)$y
  accurVal <- round(100-50*trapint(xg,abs(MCpostg-VApostg)),2)
  
  # Obtain approximate posterior plots:
  
  xlimVal <- c(xgLow,xgUpp)
  ylimVal <- range(c(VApostg,MCpostg))
  
  plot(xg,VApostg,type="n",bty="l",main=expression(paste(u[21])),
       cex.main=cex.mainVal,xlab="",ylab="approx. posterior",
       cex.lab=cex.labVal,cex.axis=cex.axisVal,ylim=ylimVal)
  lines(xg,MCpostg,lwd=lwdVal,col=colMC,lty=ltyMC)
  lines(xg,VApostg,lwd=lwdVal,col=colVA,lty=ltyVA)
  
  if(showAccuracy)
  {
    # Plot accuracy value:
    
    text(0.15*min(xlimVal)+0.85*max(xlimVal),0.1*min(ylimVal)+0.9*max(ylimVal),
         as.character(paste(accurVal,"%",sep="")),cex=cex.accur,col=colAcc)
    text(0.15*min(xlimVal)+0.85*max(xlimVal),0.25*min(ylimVal)+0.75*max(ylimVal),
         "accuracy",cex=0.5*cex.accur,col=colAcc)
  }
  
  # Add true value:
  
  lines(rep(u[2,2],2),c(0,ylimVal[2]),col=colTruth)
  abline(h=0)
  
  ### Plots for Sigma ###
  
  # Obtain Monte Carlo sample from the q-density covariance
  # matrix (interpretable) parameters:
  
  nMC <- 10000
  sigma1.q.MC <- rep(NA,nMC) ; sigma2.q.MC <- rep(NA,nMC)
  rho.q.MC <- rep(NA,nMC)
  
  for (iMC in 1:nMC)
  {
    SigmaCurr <- solve(rwish(xi.q.Sigma.VMP-q+1,solve(Lambda.q.Sigma.VMP)))
    sigma1.q.MC[iMC] <- sqrt(SigmaCurr[1,1])
    sigma2.q.MC[iMC] <- sqrt(SigmaCurr[2,2])
    rho.q.MC[iMC] <- SigmaCurr[1,2]/(sigma1.q.MC[iMC]*sigma2.q.MC[iMC])
  }
  
  # Obtain density estimates and accuracy scores:
  
  parName <- c(expression(sigma[1]),expression(sigma[2]),expression(rho))
  parTrue <- c(sqrt(SigmaTrue[1,1]),sqrt(SigmaTrue[2,2]),
               SigmaTrue[1,2]/sqrt(SigmaTrue[1,1]*SigmaTrue[2,2]))
  
  accurValVec <- rep(NA,3)
  for (iParm in 1:3)
  {
    if (iParm==1) {xVAsamp <- sigma1.q.MC  ; xMCsamp <- sigma1MCMC 
    parNameVal <- parName[1]}
    
    if (iParm==2) {xVAsamp <- sigma2.q.MC  ; xMCsamp <- sigma2MCMC 
    parNameVal <- parName[2]}
    
    if (iParm==3) {xVAsamp <- rho.q.MC  ; xMCsamp <-  rhoMCMC 
    parNameVal <- parName[3]}
    
    xgLow <- quantile(c(xVAsamp,xMCsamp),0.0001)
    xgUpp <- quantile(c(xVAsamp,xMCsamp),0.9999)
    
    xg <- seq(xgLow,xgUpp,length=gridSize) 
    
    destVA <- bkde(xVAsamp,bandwidth=dpik(xVAsamp),
                   range.x=range(xg),gridsize=gridSize)
    VApostg <- destVA$y
    
    destMC <- bkde(xMCsamp,bandwidth=dpik(xMCsamp),
                   range.x=range(xg),gridsize=gridSize)
    MCpostg <- destMC$y
    
    accurValVec[iParm] <- round(100-50*trapint(xg,abs(MCpostg-VApostg)),2)
    
    # Obtain approximate posterior plots:
    
    xlimVal <- c(xgLow,xgUpp)
    ylimVal <- range(c(VApostg,MCpostg))
    
    plot(xg,VApostg,type="n",bty="l",main=parNameVal,cex.main=cex.mainVal,xlab="",
         ylab="approx. posterior",cex.lab=cex.labVal,cex.axis=cex.axisVal,ylim=ylimVal)
    lines(xg,MCpostg,lwd=lwdVal,col=colMC,lty=ltyMC)
    lines(xg,VApostg,lwd=lwdVal,col=colVA,lty=ltyVA)
    
    if(showAccuracy)
    {
      # Plot accuracy value:
      
      text(0.15*min(xlimVal)+0.85*max(xlimVal),0.1*min(ylimVal)+0.9*max(ylimVal),
           as.character(paste(round(accurValVec[iParm]),"%",sep="")),cex=cex.accur,col=colAcc)
      text(0.15*min(xlimVal)+0.85*max(xlimVal),0.25*min(ylimVal)+0.75*max(ylimVal),
           "accuracy",cex=0.5*cex.accur,col=colAcc)
    }
    
    # Add true value:
    
    lines(rep(parTrue[iParm],2),c(0,ylimVal[2]),col=colTruth)
    abline(h=0)
  }
  
  ### Legend ###
  
  plot.new()
  
  legend("center",legend=c("VMP","MCMC"),col=c(colVA,colMC),
         lwd=c(lwdVal,lwdVal),lty=c(ltyVA,ltyMC),cex=2)

  if(doPDF) dev.off()
}

if (plotVMPfits)
{
  # Load required library:
  
  library(lattice)
  
  # Select colour for fitted value:
  
  fitCol <- "blue"
  
  # Partition mu_{q(beta,u)} and Sigma_{q(beta,u)}:
  
  mu.q.beta.VMP <- mu.q.betau.VMP[1:p]
  Sigma.q.beta.VMP <- Sigma.q.betau.VMP[1:2,1:2]
  mu.q.u.VMP <- Sigma.q.u.VMP <- E.q.beta.u.VMP <- vector("list",m)
  for (i in 1:m)
  {
    mu.q.u.VMP[[i]] <- mu.q.betau.VMP[(p+1+(i-1)*q):(p+i*q)]
    Sigma.q.u.VMP[[i]] <- Sigma.q.betau.VMP[(p+1+(i-1)*q):(p+i*q),(p+1+(i-1)*q):(p+i*q)]
    E.q.beta.u.VMP[[i]] <- Sigma.q.betau.VMP[1:p,(p+1+(i-1)*q):(p+i*q)]
  }
  
  xg <- seq(0,1,length=31)
  Xg <- cbind(1,xg)
  XgXg <- cbind(Xg,Xg)
  
  xgList <- vector("list",m)
  meangList <- vector("list",m)
  meangListTrue <- vector("list",m)
  lowgList <- vector("list",m)
  uppgList <- vector("list",m)
  
  for (i in 1:m)
  {
    xgList[[i]] <- xg
    meangList[[i]] <- (mu.q.beta.VMP[1] + mu.q.u.VMP[[i]][1] 
                       + (mu.q.beta.VMP[2] + mu.q.u.VMP[[i]][2])*xg)
    
    meangListTrue[[i]] <- (beta0True + u[i,1]) + (beta1True + u[i,2])*xg
    
    covMatFull <- matrix(0,(p+q),(p+q)) 
    
    INDSbeta <- 1:p ; INDSu <- (p+1):(p+q)
    
    covMatFull[INDSbeta,INDSbeta] <- Sigma.q.beta.VMP
    covMatFull[INDSu,INDSu] <- Sigma.q.u.VMP[[i]]
    covMatFull[INDSbeta,INDSu] <- E.q.beta.u.VMP[[i]]
    covMatFull[INDSu,INDSbeta] <- t(E.q.beta.u.VMP[[i]])
    
    sdCurr <-  sqrt(diag(XgXg%*%covMatFull%*%t(XgXg)))
    
    lowgList[[i]] <- meangList[[i]] - qnorm(0.975)*sdCurr
    uppgList[[i]] <- meangList[[i]] + qnorm(0.975)*sdCurr
  }
  
  figFits <- xyplot(y~x|as.factor(idnum),groups=idnum,
                    data=data.frame(x=x,y=y,idnum=idnum),
                    layout=layoutVec,main="",strip=FALSE,
                    xlab="x",ylab="y",
                    panel=function(x,y,subscripts,groups)
                    {
                      i <- panel.number()
                      panel.grid()
                      
                      panel.superpose(x[order(x)],y[order(x)],subscripts,groups,
                                      type="p",col="green3",pch=1,cex=0.5)
                      
                      panel.xyplot(xgList[[i]],meangListTrue[[i]],
                                   col="indianred3",type="l",lwd=2)
                      
                      panel.xyplot(xgList[[i]],meangList[[i]],
                                   col=fitCol,type="l",lwd=2)
                      
                      panel.xyplot(xgList[[i]],lowgList[[i]],
                                   col=fitCol,type="l",lwd=2,lty=2)
                      
                      panel.xyplot(xgList[[i]],uppgList[[i]],
                                   col=fitCol,type="l",lwd=2,lty=2)
                      
                    })
  
  print(figFits)
}

############ End of t2Lev ############
