########## R function: updateGaussPenFrag ##########

# For updating natural parameters of factor to stochastic
# node messages for the Gaussian penalization fragment.

# Last changed: 18 MAY 2020

updateGaussPenFrag <- function(mutheta=NULL,Sigmatheta=NULL,dtheta,etaIN)
{
   # The naming order for etaIN is:
  
   # 1. (theta0,...,thetaL)->p(theta0,...,thetaL|Theta1,...,ThetaL)
   # 2. p(theta0,...,thetaL|Theta1,...,ThetaL)->(theta0,...,thetaL)
   # 3. Theta1->p(theta0,...,thetaL|Theta1,...,ThetaL)
   # 4. p(theta0,...,thetaL|Theta1,...,ThetaL)->Theta1
   # 5. Theta2->p(theta0,...,thetaL|Theta2,...,ThetaL)
   # 6. p(theta0,...,thetaL|Theta1,...,ThetaL)->Theta2
   # .
   # .
   # .
   # (2*L+1). ThetaL->p(theta0,...,thetaL|Theta1,...,ThetaL)
   # (2*L+2). p(theta0,...,thetaL|Theta1,...,ThetaL)->ThetaL
  
   # The naming order for etaOUT is:

   # 1. "p(theta0,...,thetaL|Theta1,...,ThetaL)->(theta0,...,thetaL)"
   # 2. "p(theta0,...,thetaL|Theta1,...,ThetaL)->Theta1"
   # .
   # .
   # .
   # (L+1). "p(theta0,...,thetaL|Theta1,...,ThetaL)->ThetaL"

   # Set up is.wholenumber():
 
   is.wholenumber <- function(x,tol=sqrt(.Machine$double.eps))
      return(abs(x-round(x))<tol)
  
   # Determine value of L:

   L <- 0.5*length(etaIN) - 1

   # Allocate natural parameter lists:

   etaSUM <- vector("list",(2*L))
   etaOUT <- vector("list",(L+1))

   # Obtain the dimension values:

   dTheta <- rep(NA,L)

   if (is.null(mutheta)|is.null(Sigmatheta))
      dtheta0 <- 0

   if ((!is.null(mutheta))&(!is.null(Sigmatheta)))
     dtheta0 <- length(mutheta)

   for (ell in 1:L)
      dTheta[ell] <- sqrt(length(etaIN[[2*ell+1]][-1]))

   if (any(dTheta>2)) stop("dTheta values exceeding 2 not yet supported.")

   mVec <- dtheta/dTheta
   if (any(!is.wholenumber(mVec)))
       stop("mVec is not a vector of integers.")

   # Obtain the sums of natural parameters along each edge:
   
   etaSUM[[1]] <- etaIN[[1]] + etaIN[[2]]  
   for (ell in 1:L)
      etaSUM[[ell+1]] <-  etaIN[[2*ell+1]] + etaIN[[2*ell+2]]
   
   # Update the factor to stochastic node message natural parameters, starting
   # with the message from p(theta0,...,thetaL|Theta1,...,ThetaL) to (theta0,...,thetaL):

   pt1 <- NULL
   tmpMat <- matrix(0,(dtheta0+sum(dTheta*mVec)),(dtheta0+sum(dTheta*mVec)))
   if (dtheta0>0)
   {
      pt1 <- solve(Sigmatheta,mutheta)
      tmpMat[1:dtheta0,1:dtheta0] <- solve(Sigmatheta)
   }
   pt1 <- c(pt1,rep(0,sum(dtheta)))
      
   indStt <- dtheta0 + 1
   for (ell in 1:L)
   {

     indEnd <- indStt+dTheta[ell]*mVec[ell] - 1
     indsCurr <- indStt:indEnd

      etaCurr <- etaSUM[[ell+1]]
      OmegaCurr <- (etaCurr[1] + 0.5*(dTheta[ell]+1))*solve(vecInverse(etaCurr[-1]))
      
      if (dTheta[ell]==1)
         tmpMat[indsCurr,indsCurr] <- as.numeric(OmegaCurr)*diag(mVec[ell])

      if (dTheta[ell]==2)
         tmpMat[indsCurr,indsCurr] <- kronecker(diag(mVec[ell]),OmegaCurr)

      
      indStt <- indStt + dtheta[ell]

   }      
   pt2 <- -0.5*vec(tmpMat)
   etaOUT[[1]] <- c(pt1,pt2)
   
   # Now update the factor to stochastic node messages natural parameters
   # for the messages from p(theta0,...,thetaL|Theta1,...,ThetaL) to Theta_ell,
   # for each of 1<=ell<=L:

   aVec <- etaSUM[[1]]
   ncC <- (sqrt(4*length(aVec) + 1) - 1)/2
   a1 <- aVec[1:ncC]
   a2 <- aVec[(ncC+1):(ncC*(ncC+1))]
   vecinva2 <- vecInverse(a2)
   XiMat <- 0.25*solve(vecinva2,(t(solve(vecinva2,tcrossprod(a1)))-2*diag(ncC)))
   
   indStt <- dtheta0 + 1
   for (ell in 1:L)
   {
      indEnd <- indStt + dtheta[ell] - 1
      indsCurr <- indStt:indEnd

      if (dTheta[ell]==1)
          etaOUT[[ell+1]] <-  -0.5*c(mVec[ell],sum(diag(XiMat[indsCurr,indsCurr])))
        
      if (dTheta[ell]==2)
      {
         indMat <- vector("list",4)
         oddInds <- seq(indStt,(indEnd-1),by=2)
         eveInds <- oddInds + 1
         indMat[[1]] <- cbind(oddInds,oddInds)
         indMat[[2]] <- cbind(eveInds,oddInds)
         indMat[[3]] <- cbind(oddInds,eveInds)
         indMat[[4]] <- cbind(eveInds,eveInds)

         vecTMP <- NULL
         for (i in 1:4)
             vecTMP <- c(vecTMP,sum(XiMat[indMat[[i]]]))

         etaOUT[[ell+1]] <- -0.5*c(mVec[ell],vecTMP)
      }
        
      indStt <- indStt + dtheta[ell] 
   }

   return(etaOUT)
}

############ End of updateGaussPenFrag ############
