########## R function: updateIterInvGWishFrag ##########

# For the updating natural parameters of factor to stochastic
# node messages for the Iterated Inverse G-Wishart fragment.
# This function corresponds to Algorithm 2 of 
# Maestrini & Wand (2021).

# Last changed: 08 OCT 2021

updateIterInvGWishFrag <- function(xi,G,GetaIN)
{  
   require("matrixcalc")

   # The only component of GetaIN$G corresponds to:

   # 1. A->p(Sigma|A)

   # The naming order for GetaIN$eta is:
  
   # 1. Sigma->p(Sigma|A)
   # 2. p(Sigma|A)->Sigma
   # 3. A->p(Sigma|A)
   # 4. p(Sigma|A)->A

   # The naming order for GetaOUT is:

   # 1. p(Sigma|A)->Sigma
   # 2. p(Sigma|A)->A

   # Break up GetaIN into graph and eta vector components:

   GIN <- GetaIN$G
   etaIN <- GetaIN$eta

   # Set up is.wholenumber():
 
   is.wholenumber <- function(x,tol=sqrt(.Machine$double.eps))
      return(abs(x-round(x))<tol)
 
   # Obtain the dimension value:

   lenVal <- length(GetaIN$eta[[2]]) - 1
   dmn <- (sqrt(8*lenVal +1) - 1)/2

   if(!is.wholenumber(dmn)) stop("dmn is not an integer.")
   
   # Obtain the required duplication matrix and its 
   # Moore-Penrose inverse:

   if (dmn==1) Dd <- 1
   if (dmn>1) Dd <- duplication.matrix(dmn)

   DdPlus <- solve(crossprod(Dd),t(Dd))

   # Do the initial graph updates:

   GpSigmaGivenAtoSigma <- G
   GpSigmaGivenAtoA <- GIN

   # Obtain the sums of natural parameters along each edge:

   etapSigmaGivenAconnSigma <- etaIN[[1]] + etaIN[[2]]
   etapSigmaGivenAconnA  <- etaIN[[3]] + etaIN[[4]]
 
   # Obtain omega_1 value:

   if (GpSigmaGivenAtoA=="Gfull") omega1 <- (dmn+1)/2
   if (GpSigmaGivenAtoA=="Gdiag") omega1 <- 1

   # Obtain E_q(A^{-1}):

   E.q.A.inv <- ((etapSigmaGivenAconnA[1] + omega1)
                *solve(vecInverse(crossprod(DdPlus,etapSigmaGivenAconnA[-1]))))

   if ((GpSigmaGivenAtoSigma=="Gdiag") & (dmn>1)) E.q.A.inv <- diag(diag(E.q.A.inv))

   # Obtain omega_2 value:

   if (GpSigmaGivenAtoSigma=="Gfull") omega2 <- (dmn+1)/2
   if (GpSigmaGivenAtoSigma=="Gdiag") omega2 <- 1

   # Obtain E_q(Sigma^{-1}):

   E.q.Sigma.inv <- ((etapSigmaGivenAconnSigma[1] + omega2)
                     *solve(vecInverse(crossprod(DdPlus,etapSigmaGivenAconnSigma[-1]))))

   if ((GpSigmaGivenAtoA=="Gdiag") & (dmn>1)) E.q.Sigma.inv <- diag(diag(E.q.Sigma.inv))

   # Prepare the output list:

   GetaOUT <- list(G = list(GpSigmaGivenAtoSigma,GpSigmaGivenAtoA),
                   eta = list(-0.5*c((xi+2),crossprod(Dd,vec(E.q.A.inv))),
                              -0.5*c((xi+2-2*omega2),crossprod(Dd,vec(E.q.Sigma.inv)))))

   # Return the output list:
   
   return(GetaOUT)
}

############ End of updateIterInvGWishFrag ############
